--[[ *** LOAD LIBRARIES *** 
	Force used libraries to load. This is necessary because load on demand addons (in our case,
	disembedded libs) will no longer load even when listed in OptionalDeps and can only be loaded
	via LoadAddOn(). This must be done before a LibStub("AceAddon-3.0"):NewAddon() call due to
	ADDON_LOADED firing, as AceAddon-3.0 will call :OnInit() on any addon's ADDON_LOADED event due
	to needing to support submodules which may not be in their own addons.
]]--
LoadAddOn("Ace3")
LoadAddOn("LibDataBroker-1.1")
LoadAddOn("LibDBIcon-1.0")
LoadAddOn("LibQTip-1.0")

--[[ *** ADDON DECLARATION *** ]]--
local modName = "Guilder"
local _G = _G
local Guilder = LibStub("AceAddon-3.0"):NewAddon("Guilder", "AceComm-3.0", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0")
_G.Guilder = Guilder

--[[ *** DEFINE LIBRARIES *** ]]--
local ACD3 = LibStub("AceConfigDialog-3.0")
-- load localisation library
local L = LibStub("AceLocale-3.0"):GetLocale("Guilder")
Guilder.L = L
-- load ChatThrottleLib - used to make sure you don't get disconnected due to high chat volume
local CTL = assert(ChatThrottleLib, "Guilder requires ChatThrottleLib")
-- load LibDataBroker and LibDBIcon
local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)
LG = LibStub("LibGuild-1.0", true)
if not LG then
	Guilder:Print("LibGuild-1.0 not found!")
	return
end
LP = LibStub("LibParty-1.0", true)
if not LP then
	Guilder:Print("LibParty-1.0 not found!")
	return
end
LR = LibStub("LibRaid-1.0", true)
if not LR then
	Guilder:Print("LibRaid-1.0 not found!")
	return
end
-- load QTip library for tooltips
local QTC = LibStub("LibQTip-1.0")

--[[ *** SET UP VARIABLES *** ]]--

-- make global API calls local to speed things up
local pairs = _G.pairs
local string = _G.string
local table = _G.table
local CreateFrame = _G.CreateFrame
local GetFriendInfo = _G.GetFriendInfo
local GetNumFriends = _G.GetNumFriends
local IsInGuild = _G.IsInGuild
local UIParent = _G.UIParent

-- colours stored for code readability
RESET = "|r"
BLIZZARD = "|cFF00C0FF"
BLUE = "|cFF0070DD"
CYAN = "|cFF00C0FF"
GREEN = "|cFF1EFF00"
GREY = "|cFF999999"
ORANGE = "|cFFFF8400"
PURPLE = "|cFF9F3FFF"
RED = "|cFFFF0000"
WHITE = "|cFFFFFFFF"
YELLOW = "|cFFFFFF00"

-- Create an LDB object
if LDB then
	LDBObj = LDB:NewDataObject("Guilder", {
		type = "launcher",
		label = "Guilder",
		text = "Guilder",
		icon = "Interface\\AddOns\\Guilder\\Images\\icon.tga"
	})
end

-- if the LibDBIcon library is available, create a minimap icon
local f = CreateFrame("Frame")
f:SetScript("OnEvent", function()
	if LDBIcon then
		LDBIcon:Register("Guilder", LDBObj, Guilder.db.profile.minimap)
	end
	f:UnregisterEvent("PLAYER_LOGIN")
end)
f:RegisterEvent("PLAYER_LOGIN")

-- Guilder variables
msgFrame = nil
msgFrameCreated = "0"
myName = UnitName("player")
myFaction, _ = UnitFactionGroup("player")
myRealm = GetRealmName()
myNameRealm = myName.." - "..myRealm
myVersionNumber = GetAddOnMetadata("Guilder", "Version")
local gldrTooltip = nil

--[[ *** DEFAULT SETTINGS *** ]]--

local defaults = {
	profile = {
		info = {
			showOnline = false,
			showTotal = false,
			showOffline = true,
			showModules = false,
			showHint = true,
		},
		minimap = {
			hide = false,
			minimapPos = 90,
			radius = 80,
		},
		modules = {},
		tooltip = {},
		dbCache = {},
	},
}

--[[ *** CONFIGURATION MENU *** ]]--

local options = {
	type = "group",
	args = {
		defaultArgs = {
			type = "group",
			name = L["Guilder"],
			args = {
				header = {
					type = "header",
					name = GetAddOnMetadata("Guilder", "Title"),
					order = 9,
				},
				Info = {
					type = "description",
					name = GetAddOnMetadata("Guilder", "Notes"),
					order = 10,
				},
				settingsHeader = {
					type = "header",
					name = L["Settings"],
					order = 11,
				},
				aceConfig = {
					type = "execute",
					name = L["Standalone Config"],
					desc = L["Open a standalone config window. You might consider installing |cffffff00BetterBlizzOptions|r to make the Blizzard UI options panel resizable."],
					order = 12,
					func = function()
						InterfaceOptionsFrame:Hide()
						ACD3:SetDefaultSize("Guilder", 500, 550)
						ACD3:Open("Guilder")
					end,
				},
				info = {
					type = "group",
					name = L["Information"],
					order = 20,
					guiInline = true,
					args = {
						Online = {
							type = "toggle",
							name = L["Show Online"],
							desc = L["Show number of officers online"],
							order = 1,
							get = function() return Guilder.db.profile.info.showOnline end,
							set = function(info, v) Guilder.db.profile.info.showOnline = v; Guilder:UpdateText() end,
						},
						Total = {
							type = "toggle",
							name = L["Show Total"],
							desc = L["Show total number of officers"],
							order = 2,
							get = function() return Guilder.db.profile.info.showTotal end,
							set = function(info, v) Guilder.db.profile.info.showTotal = v; Guilder:UpdateText() end,
						},
						Offline = {
							type = "toggle",
							name = L["Show Offline Officers"],
							desc = L["Show offline officers in tooltip"],
							order = 3,
							get = function() return Guilder.db.profile.info.showOffline end,
							set = function(info, v) Guilder.db.profile.info.showOffline = v end,
						},
						Modules = {
							type = "toggle",
							name = L["Show Module State"],
							desc = L["Show module state in tooltip"],
							order = 4,
							get = function() return Guilder.db.profile.info.showModules end,
							set = function(info, v) Guilder.db.profile.info.showModules = v end,
						},
						Hint = {
							type = "toggle",
							name = L["Show Hint"],
							desc = L["Show hint at the bottom of the tooltip"],
							order = 5,
							get = function() return Guilder.db.profile.info.showHint end,
							set = function(info, v) Guilder.db.profile.info.showHint = v end,
						},
						Minimap = {
							type = "toggle",
							name = L["Hide Minimap Icon"],
							desc = L["Hides the minimap icon"],
							order = 6,
							get = function () return Guilder.db.profile.minimap.hide end,
							set = function (info, v)
								Guilder.db.profile.minimap.hide = v
								if v then
									LDBIcon:Hide("Guilder")
								else
									LDBIcon:Show("Guilder")
								end
							end,
						}
					},
				},
				ranks = {
					type = "group",
					name = L["Ranks"],
					order = 21,
					guiInline = true,
					args = {
						info = {
							type = "description",
							name = L["You are not in a guild!"],
						},
					},
				},
				rescan = {
					type = "group",
					name = L["Miscellaneous"],
					order = 22,
					guiInline = true,
					args = {
						info = {
							type = "description",
							name = L["Things don't always go as planned. If you can't see the guild ranks and you are definitely in a guild, please use the rescan button."],
							order = 1,
						},
						rescan = {
							type = "execute",
							name = L["Rescan"],
							order = 2,
							func = function() Guilder:CreateRanks() end,
						},
					}
				}
			},
		},
		config = {
			type = "execute",
			guiHidden = true,
			name = L["Configure"],
			desc = L["Configure"],
			func = Guilder.OpenConfig
		},
		modules = {
			type = "group",
			name = L["Modules"],
			desc = L["Modules"],
			args = {}
		},
	},
}

--[[ *** EVENT FUNCTIONS *** ]]--

Guilder:SetDefaultModuleState(false)

do
	local timer, t = nil, 0
	local function update()
		t = t + arg1
		if t > 0.5 then
			timer:SetScript("OnUpdate", nil)
			Guilder:UpdateConfig()
		end
	end
	function Guilder:SetUpdateConfig()
		t = 0
		timer = timer or CreateFrame("Frame", nil, UIParent)
		timer:SetScript("OnUpdate", update)
	end
end

-- this function is part of the Blizzard API
function Guilder:OnDisable()
	self:UnregisterAllEvents()
	LG:UnregisterAllEvents()
end

-- this function is part of the Blizzard API
function Guilder:OnEnable()
	-- create the generic message frame
	self:CreateMsgFrame()
	if not self.db.profile.welcomeShown then
		Guilder:Print(L["Welcome to Guilder! Type /guilder or /gldr to configure."])
		self.db.profile.welcomeShown = true
	end
	-- load modules
	for k, v in self:IterateModules() do
		if self.db.profile.modules[k] ~= false then
			v:Enable()
		else
			v:Disable()
		end
	end
	-- if we are in a guild, register a callback to update the LDB text
	if IsInGuild() then
		LG:RegisterCallback("Update", function() Guilder:UpdateText() end)
		-- update the ranks if there is an update on the guild roster
		self:RegisterEvent("GUILD_ROSTER_UPDATE")
		-- Update the ranks if the player leaves / joins a guild
		self:RegisterEvent("PLAYER_GUILD_UPDATE")
	end
	-- Catch whispers and filter what should be shown
	-- TODO: stop WIM from showing everyting
	self:RawHook("ChatFrame_MessageEventHandler", "WhisperHandler", true)
	self:RegisterEvent("CHAT_MSG_WHISPER")
end

function Guilder:GetOptions()
	return options
end
-- this function is part of the Blizzard API
function Guilder:OnInitialize()
	-- Register database
	local db = LibStub("AceDB-3.0"):New("GuilderDB", defaults, "Default")
	self.db = db
	-- register options wit the Blizzard interface
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Guilder", options)
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("GuilderModules", options.args.modules)
	self.optionsFrame = ACD3:AddToBlizOptions("Guilder", GetAddOnMetadata("Guilder", "Title"), nil, "defaultArgs")

	-- fill up our modular options...
	for k, v in self:IterateModules() do
		options.args.modules.args[k:gsub(" ", "_")] = {
			type = "group",
			name = (v.modName or k),
			args = nil
		}
		local t
		if v.GetOptions then
			t = v:GetOptions()
			t.settingsHeader = {
				type = "header",
				name = L["Settings"],
				order = 12
			}		
		end
		t = t or {}
		t.toggle = {
			type = "toggle", 
			name = v.toggleLabel or (L["Enable "] .. (v.modName or k)), 
			width = "double",
			desc = v.Info and v:Info() or (L["Enable "] .. (v.modName or k)), 
			order = 11,
			get = function()
				return Guilder.db.profile.modules[k] ~= false or false
			end,
			set = function(info, v)
				Guilder.db.profile.modules[k] = v
				if v then
					Guilder:EnableModule(k)
					Guilder:Print(L["Enabled"], BLIZZARD..k..RESET, L["Module"])
				else
					Guilder:DisableModule(k)
					Guilder:Print(L["Disabled"], BLIZZARD..k..RESET, L["Module"])
				end
			end
		}
		t.header = {
			type = "header",
			name = v.modName or k,
			order = 9
		}
		if v.Info then
			t.description = {
				type = "description",
				name = v:Info() .. "\n\n",
				order = 10
			}
		end
		options.args.modules.args[k:gsub(" ", "_")].args = t
	end	
	-- sort modules alphabetically before adding to the options
	local moduleList = {}
	local moduleNames = {}
	for k, v in pairs(options.args.modules.args) do
		moduleList[v.name] = k
		tinsert(moduleNames, v.name)
	end
	table.sort(moduleNames)
	for _, name in ipairs(moduleNames) do
		ACD3:AddToBlizOptions("GuilderModules", name, "Guilder", moduleList[name])
	end
	-- expose the modules section for modules
	Guilder.ModuleOptions = options.args.modules.args
	-- add the Profiles and About options
	if not options.args.Profiles then
 		options.args.Profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(db)
		self.optionsFrame.Profiles = ACD3:AddToBlizOptions("Guilder", L["Profiles"], "Guilder", "Profiles")
		-- Add in the about panel to the Bliz options (but not the ace3 config)
		if LibStub:GetLibrary("LibAboutPanel", true) then
			self.optionsFrame["About"] = LibStub:GetLibrary("LibAboutPanel").new("Guilder", "Guilder")
		else
			Guilder:Print(RED.."LibAboutPanel not loaded.")
		end
	end
	-- Register slash commands
	self:RegisterChatCommand("gldr", "OpenConfig")
	self:RegisterChatCommand("guilder", "OpenConfig")
	-- TODO: hook Profile functions -- not used atm
	-- db.RegisterCallback(self, "OnProfileChanged", "SetUpdateConfig")
	-- db.RegisterCallback(self, "OnProfileCopied", "SetUpdateConfig")
	-- db.RegisterCallback(self, "OnProfileReset", "SetUpdateConfig")
	-- first time this character is using the addon, so create and initialise rank tables
	if self.db.profile.dbCache[myNameRealm] == nil then
		self.db.profile.dbCache[myNameRealm] = {}
		self.db.profile.dbCache[myNameRealm].ranknames = {}
		self.db.profile.dbCache[myNameRealm].ranks = {}
		for i = 1, 10 do
			self.db.profile.dbCache[myNameRealm].ranknames[i] = ""
			self.db.profile.dbCache[myNameRealm].ranks[i] = "3-Member"
		end
	end
	-- create alias for cached information
	dbCache = self.db.profile.dbCache[myNameRealm]
	-- if character is in a guild, get the rank names for the options frame
	if IsInGuild then
		GuildRoster()
		self:CreateRanks(dbCache.ranknames[1] ~= "")
	end
end

--[[ *** CREATE MESSAGE FRAME *** ]]--

--- This function creates a frame to use as a popup window to share information
-- @usage internal use only
function Guilder:CreateMsgFrame()
--[[
	This function was adapted from the similar function provided in the ACE library API for
	creating an 'About' information window for a FuBar compatible plugin mod.
]]
	-- create the window frame object and configure its properties.
	msgFrame = CreateFrame("Frame", "Acemod30msgFrame", UIParent, "DialogBoxFrame")
	msgFrame:SetWidth(500)
	msgFrame:SetHeight(400)
	msgFrame:SetPoint("CENTER")
	msgFrame:SetBackdrop({
		bgFile = [[Interface\DialogFrame\UI-DialogBox-Background]],
		edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
		tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 5, right = 5, top = 5, bottom = 5 }
	})
	msgFrame:SetBackdropColor(0,0,0,1)
	-- set the window's title text.
	local text = msgFrame:CreateFontString(nil, "OVERLAY", "GameFontHighlightLarge")
	msgFrame.title = text
	text:SetPoint("TOP", 0, -5)
	-- initialise the window's text arrays.
	msgFrame.lefts = {}
	msgFrame.rights = {}
	msgFrame.textLefts = {}
	msgFrame.textRights = {}
	
	--- This function clears the message popup window
	-- @param nil
	-- @usage msgFrame:Clear()
	-- @returns nil
	function msgFrame:Clear()
		self.title:SetText("")
		for i = 1, #self.lefts do
			self.lefts[i] = nil
			self.rights[i] = nil
		end
	end

	--- This function prepares the popup window for a new version check
	-- @param nil
	-- @usage msgFrame:PrepareForVersionCheck()
	-- @returns nil
	function msgFrame:PrepareForVersionCheck()
		self:Clear()
		msgFrame.title:SetText("Guilder")
		msgFrame:AddLine(L["Player"], L["Version"])
		msgFrame:AddLine("@NULL@", " ")
		msgFrame:AddLine(myName, BLIZZARD..myVersionNumber)
	end

	--- This function adds a line of text to the popup window
	-- @param left message to display on the left half of the line (use "@NULL@" do create an empty line)
	-- @param right message to display on the right half of the line
	-- @usage msgFrame:Addline("Left Text", "Right Text")
	-- @returns nil
	function msgFrame:AddLine(left, right)
		msgFrame.lefts[#msgFrame.lefts+1] = left
		msgFrame.rights[#msgFrame.rights+1] = right
	end

	-- show the window
	local msgFrame_Show = msgFrame.Show

	--- This function displays the popup window
	-- @param nil
	-- @usage msgFrame:Show()
	-- @returns nil
	function msgFrame:Show(...)
		local maxLeftWidth = 0
		local maxRightWidth = 0
		local textHeight = 0
		for i = 1, #self.lefts do
			if not self.textLefts[i] then
				local left = msgFrame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
				self.textLefts[i] = left
				local right = msgFrame:CreateFontString(nil, "OVERLAY", "GameFontHighlight")
				self.textRights[i] = right
				if i == 1 then
					left:SetPoint("TOPRIGHT", msgFrame, "TOPLEFT", 75, -35)
				else
					left:SetPoint("TOPRIGHT", self.textLefts[i-1], "BOTTOMRIGHT", 0, -5)
				end
				right:SetPoint("LEFT", left, "RIGHT", 5, 0)
			end
			if self.lefts[i] == "@NULL@" then
				self.textLefts[i]:SetText(" ")
			else
				self.textLefts[i]:SetText(self.lefts[i] .. ":")
			end

			self.textRights[i]:SetText(self.rights[i])
			local leftWidth = self.textLefts[i]:GetWidth()
			local rightWidth = self.textRights[i]:GetWidth()
			textHeight = self.textLefts[i]:GetHeight()
			if maxLeftWidth < leftWidth then
				maxLeftWidth = leftWidth
			end
			if maxRightWidth < rightWidth then
				maxRightWidth = rightWidth
			end
		end
		for i = #self.lefts+1, #self.textLefts do
			self.textLefts[i]:SetText('')
			self.textRights[i]:SetText('')
		end
		msgFrame:SetWidth(75 + maxRightWidth + 20)
		msgFrame:SetHeight(#self.lefts * (textHeight + 5) + 100)

		msgFrame_Show(self, ...)
	end
	-- Set a flag indicating that the window has been created.
	msgFrameCreated = "1"
end

--[[ *** GENERIC FUNCTIONS *** ]]--

--- This function receives a string and converts it into a table of arguments, using "separator" as the separator character
-- @param message the input string
-- @param separator the separator character
-- @usage Guilder:GetArgs(arg1, " ")
-- @returns table of arguments
function Guilder:GetArgs(message, separator)
	if message == nil then
		message = ""
	end

	local args = {}
	local i = 0
	-- search for seperators in the string and return the separated data
	for value in string.gmatch(message, "[^"..separator.."]+") do
		i = i + 1
		args[i] = string.lower(value)
	end
	return args
end

--- This function dynamically creates the select boxes for each rank in the guild using the rank name as the name
-- @param useCache uses a cached copy of the guild information if true, otherwise rebuilds it
-- @usage CreateRanks(true)
-- @usage CreateRanks()
-- @return nil
function Guilder:CreateRanks(useCache)
	if IsInGuild() then
		-- get rank names
		local r = {}
		for i = 1, GuildControlGetNumRanks() do
			local rankName
			-- if useCache = true then used the cached ranks, else recreate them
			if useCache then
				rankName = dbCache.ranknames[i]
			else
				rankName = GuildControlGetRankName(i)
				if rankName == nil or rankName == "" then rankName = "Rank "..i end
				if dbCache.ranknames == nil then
					dbCache.ranknames = {}
				end
				dbCache.ranknames[i] = rankName
			end
			-- create the options
			r["rank"..i] = {
			  	type = "select",
			  	name = rankName,
				desc = L["Select the rank type for this rank"],
				order = i,
				values = { ["1-Officer"] = L["Officer"], ["2-OffAlt"] = L["Officer Alt"], ["3-Member"] = L["Member"], ["4-Alt"] = L["Alt"] },
				get = function() return Guilder.db.profile.dbCache[myNameRealm].ranks[i] end,
				set = function(info, v)
					Guilder.db.profile.dbCache[myNameRealm].ranks[i] = v
					Guilder:UpdateText()
				end,
			}
		end
		-- place rank names in the config screen
		options.args.defaultArgs.args.ranks.args = r
	else
		-- user not in a guild so display a message instead
		options.args.defaultArgs.args.ranks.args = {
			info = {
				type = "description",
				name = L["You are not in a guild!"],
			},
		}
	end
end

--[[ *** CATCH EVENTS *** ]]--

--- This function checks whispers to filter out those the player should not see
-- @usage internal use only
function Guilder:WhisperHandler(...)
	if arg1 and arg1 ~= "" then
		-- this checks incoming whispers
		msg = arg1
		if (string.find(msg, "^guilder") and event == "CHAT_MSG_WHISPER") then
			return
		end
		-- this checks outgoing whispers
		if (string.find(msg:lower(), "^%[guilder%]") and event == "CHAT_MSG_WHISPER_INFORM") then
			return
		end
	end
	-- if we get here, we pass the whisper on to the default message handler
	return self.hooks["ChatFrame_MessageEventHandler"](...)
end

--- This function checks for a whisper starting with "gldr" or "guilder" and responds with the relevant information
-- @usage internal use only
function Guilder:CHAT_MSG_WHISPER()
	local msg = self:GetArgs(arg1, " ")
	local sender = arg2
	-- check if the sender issued a 'guilder' command.
	if msg[1] ~= "guilder" and msg[1] ~= "gldr" then
		-- ignore messages from the addon
		if msg[1] == "[guilder]" then
			return true
		end
		-- if Busy Responder module is active, forward it to the responder
		if self.db.profile.modules["Busy Responder"] then
			self:GetModule("Busy Responder"):SendBusyResponder(sender, arg1)
		end
		return true
	end
	-- process the 'guilder' command
	local command
	if msg[2] then
		command = msg[2]
	end
	-- sender requested guilder command "help"
	if command == nil or command == string.lower(L["help"]) then
		self:SendCommandHelp(sender)
		return true
	end
	-- sender requested guild information
	if command == string.lower(L["guildinfo"]) then
		self:SendGuildInfo(sender)
		return true
	end
	-- check module commands
	for k, v in self:IterateModules() do
		if command == string.lower(v:GetCommand()) then
			if self.db.profile.modules[k] ~= false then
				v:DoCommand(sender, msg)
			else
				self:SendMessage(k..": "..L["Module is not enabled."], "WHISPER", sender)
			end
			return true
		end
	end
	-- if we haven't returned yet, command must have been unrecognised.
	self:SendMessage(L["Unrecognised command received"], "WHISPER", sender)
	self:SendMessage(L["Whisper 'guilder help' for a list of valid commands"], "WHISPER", sender)
end

--- This function updates the ranks if there has been a guild roster update
-- @usage internal use only
function Guilder:GUILD_ROSTER_UPDATE()
	-- test if the ranks have changed
	local ranksChanged = false
	for i = 1, 10 do
		if dbCache.ranknames[i] ~= GuildControlGetRankName(i) then
			ranksChanged = true
			break  -- no need to test more since one change means a full rescan
		end
	end
	-- if a rank has changed, rescan the ranks
	if ranksChanged then self:CreateRanks() end
end

--- This function updates the ranks if the player joins or leaves a guild
-- @usage internal use only
function Guilder:PLAYER_GUILD_UPDATE()
	if IsInGuild() then
		-- we joined a guild, get the ranks
		self:CreateRanks()
	else
		-- we left a guild, clear the cached ranks
		for i = 1, 10 do
			dbCache.ranknames[i] = "3-Member"
		end
	end
end

--[[ *** GUILDER FUNCTIONS *** ]]--

--- This function attempts to execute the supplied argument as though it were a command entered by the user in the chat window
-- @param arg the text the user would have typed
-- @usage Guilder:ExecuteChatCommand("/silly")
-- @returns nil
function Guilder:ExecuteChatCommand(arg)
	if (not DEFAULT_CHAT_FRAME) then return end
	DEFAULT_CHAT_FRAME.editBox:SetText(arg);
	ChatEdit_SendText(DEFAULT_CHAT_FRAME.editBox);
end

--- This function attempts to find the officer's name in the notes or officer's notes
-- @param mainName name of the main character
-- @usage Guilder:FindOfficerAlt("Player")
-- @returns "AltName (MainName)"
function Guilder:FindOfficerAlt(mainName)
	if self:IsOfficer(mainName) and (not LG:IsMemberOnline(mainName)) then
		for name in LG:GetIterator("NAME", true) do
			oNote = LG:GetOfficerNote(name)
			if oNote ~= nil then
				if (LG:IsMemberOnline(name) and self:IsOfficerAlt(name) and string.match(oNote:upper(), mainName:upper())) then
					return mainName.." ("..name..")"
				end
			end
		end
	end
	return nil
end

--- This function looks for characters holding the Officer rank and then simply performs a count to determine how many such characters there
-- @param nil
-- @usage Guilder:GetNumberOfGuildOfficers
-- @returns the number of officers
function Guilder:GetNumberOfGuildOfficers()
	local numOfficers = 0
	if IsInGuild() then
		for name in LG:GetIterator("NAME", true) do
			if self:IsOfficer(name) then numOfficers = numOfficers + 1 end
		end
	end
	return numOfficers
end

--- This function looks for characters holding the Officer rank and then simply performs a count to determine how many are online at the moment
-- @param nil
-- @usage Guilder:GetNumberOfGuildOfficersOnline()
-- @returns the number of officers
function Guilder:GetNumberOfGuildOfficersOnline()
	local numOfficers = 0
	if IsInGuild() then
		for name in LG:GetIterator("NAME", true) do
			if (self:IsOfficer(name) or self:IsOfficerAlt(name)) and LG:IsMemberOnline(name) then numOfficers = numOfficers + 1 end
		end
	end
	return numOfficers
end

--- This function checks whether a character holds the ranks of officer
-- @param name name of the character
-- @usage Guilder:IsOfficer("Player")
-- @returns true if the character is an officer, otherwise returns false
function Guilder:IsOfficer(name)
	return dbCache.ranks[LG:GetRankIndex(name) + 1] == "1-Officer"
end

--- This function checks whether a character holds the rank of officer's alt
-- @param name name of the character
-- @usage Guilder:IsOfficerAlt("Player")
-- @returns true if the character is an officer, otherwise returns false
function Guilder:IsOfficerAlt(name)
	return dbCache.ranks[LG:GetRankIndex(name) + 1] == "2-OffAlt"
end

--- This function opens the Blizzard or the Ace3 configuration window
-- @param useAce (boolean) if true, open the Ace3 configuration window, otherwise open the Blizzard configuration window
-- @usage Guilder:OpenConfig(true)
-- @usage Guilder:OpenConfig()
-- @returns nil
function Guilder:OpenConfig(useAce)
	if useAce or not InterfaceOptionsFrame:IsResizable() then
		options.args.defaultArgs.guiHidden = true
		InterfaceOptionsFrame:Hide()
		ACD3:SetDefaultSize("Guilder", 500, 550)
		ACD3:Open("Guilder")
	else
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame.Profiles)
		options.args.defaultArgs.guiHidden = false
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
	end
end

--- This function sends a list of the available whisper-based commands currently recognised by the mod
-- @param recipient person who requested the help
-- @usage for internal use only
-- @returns nil
function Guilder:SendCommandHelp(recipient)
	self:SendMessage(L["Whisper 'guilder [option]' where option is one of..."], "WHISPER", recipient)
	self:SendMessage(L["[help] - displays this help"], "WHISPER", recipient)
	self:SendMessage(L["[guildinfo] - displays more information about the guild"], "WHISPER", recipient)
	self:SendMessage(L["Active modules"], "WHISPER", recipient)
	local commandList = {}
	for k, v in self:IterateModules() do
		if self.db.profile.modules[k] ~= false and v:GetHelp() then
			table.insert(commandList, v:GetHelp())
		end
	end
	if #commandList == 0 then
		self:SendMessage(L["No active modules found"], "WHISPER", recipient)
	else
		table.sort(commandList)
		for i = 1, #commandList do
			self:SendMessage(commandList[i], "WHISPER", recipient)
		end
	end
end

--- This function sends guild information
-- @param recipient person who requested the help
-- @usage for internal use only
-- @returns nil
function Guilder:SendGuildInfo(recipient)
	self:SendMessage(L["guildinfo1"], "WHISPER", recipient)
	self:SendMessage(L["guildinfo2"], "WHISPER", recipient)
	self:SendMessage(L["guildinfo3"], "WHISPER", recipient)
	self:SendMessage(L["guildinfo4"], "WHISPER", recipient)
	self:SendMessage(L["guildinfo5"], "WHISPER", recipient)
end

--- This function is ostensibly a wrapper to the normal SendChatMessage function with the following changes:
--		1: Language is checked because it may not be passed in.
--		2: If channelName = "SELF", output is routed to Guilder:Print() instead.
--		3: If channelName = "NULL", output is muted.
--	The intention for the existence of the NULL option for the channel is to allow for functions to pass a parameter that can be switched between a normal channel and null to turn on/off debug mode features.
-- @param messageText the text to send to the recipient
-- @param channelName the name of the channel on which to send it
-- @param recipient person who requested the message
-- @usage Guilder:SendMessage("Test", "SELF", "Player")
-- @returns nil
function Guilder:SendMessage(messageText, channelName, recipient)
	-- check whether the language option has been provided.
	if recipient == nil or recipient == " " then
		recipient = nil
	end
	-- can't send a whisper if there is no-one to receive it
	if channelName == "WHISPER" and recipient == nil then
		self:Print("Error! Message sent with no recipient")
		return
	end
	-- send chat output to 'chat' or to intended channel if not 'NULL'.
	if channelName == "SELF" then
		self:Print(messageText)
	elseif channelName ~= "NULL" then
		-- if it is a whisper, prepend "[Guilder] " to the message
		if channelName == "WHISPER" then messageText = "[Guilder] "..messageText end
		-- send the chatmessage using ChatThrottleLib
		CTL:SendChatMessage("BULK", "", messageText, channelName, nil, recipient)
	end
end

--- This function checks whether the named character is a friend of the person running the mod
-- @param name name of the character
-- @usage Guilder:UnitInFriends("Player")
-- @returns true if the character is in the friends list, otherwise returns false
function Guilder:UnitInFriends(name)
	local numFriends = GetNumFriends()
	for i = 1, numFriends do
		local cname, level, class, zone, online, status, note = GetFriendInfo(i)
		if online and cname == name then
			return true
		end
	end
	return false
end

--- This function refreshes the modules
-- @usage for internal use only
function Guilder:UpdateConfig()
	for k, v in self:IterateModules() do
		if v:IsEnabled() then
			v:Disable()
			v:Enable()
		end
	end
end

--- This function updates the DataBroker text
-- @param nil
-- @usage Guilder:UpdateText
-- @returns nil
function Guilder:UpdateText()
	if IsInGuild() then
		if self.db.profile.info.showOnline or self.db.profile.info.showTotal then
			local temptext = ""
			local needslash = false
			if self.db.profile.info.showOnline then
				temptext = temptext..self:GetNumberOfGuildOfficersOnline()
				needslash = true
			end
			if self.db.profile.info.showTotal then
				if needslash then temptext = temptext.."/" end
				temptext = temptext..self:GetNumberOfGuildOfficers()
			end
			if self.db.profile.info.showOnline then
				temptext = temptext.." "..L["Online"]
			else
				temptext = temptext.." "..L["Officers"]
			end
			LDBObj.text = temptext
		else
			LDBObj.text = "Guilder"
		end
	else
		if self.db.profile.info.showOnline or self.db.profile.info.showTotal then
			LDBObj.text = L["No Guild"]
		else
			LDBObj.text = "Guilder"
		end
	end
end

--[[ *** LIBDATABROKER EVENT FUNCTIONS *** ]]--

--- This function handles mouse clicks on the DataBroker object
-- @usage for internal use only
function LDBObj:OnClick(button, modifier)
	if button == "LeftButton" then
		if IsAltKeyDown() and IsControlKeyDown() then
			-- Do nothing.
		elseif IsAltKeyDown() then
			Guilder:ExecuteChatCommand("/rbs report")
			return
		elseif IsControlKeyDown() then
			Guilder.db.profile.modules["Busy Responder"] = not Guilder.db.profile.modules["Busy Responder"]
			if Guilder.db.profile.modules["Busy Responder"] then
				Guilder:EnableModule("Busy Responder")
				Guilder:Print(L["Enabled"], "Busy Responder", L["Module"])
			else
				Guilder:DisableModule("Busy Responder")
				Guilder:Print(L["Disabled"], "Busy Responder", L["Module"])
			end
			return
		end
		Guilder:ExecuteChatCommand("/rbs toggle")
	elseif button == "RightButton" then
		Guilder:OpenConfig()
	end
end

--- This function shows the tooltip if the mouse cursor hovers over the DataBroker object
-- @usage for internal use only
function LDBObj:OnEnter()
	-- define header font size and color
	local headerFont = CreateFont("gldrHeaderFont")
	headerFont:SetFont(GameTooltipText:GetFont(), 18)
	headerFont:SetTextColor(1, 1, 0)
	-- create a QTip tooltip object (will be discarded after use)
	gldrTooltip = QTC:Acquire("GuilderTooltip", 1)
	gldrTooltip:Clear()
	gldrTooltip:SmartAnchorTo(self)
	-- modify tooltip to use Blizzard tooltip settings so that tooltip mods can modify it
	gldrTooltip:SetBackdrop(GameTooltip:GetBackdrop())
	gldrTooltip:SetBackdropBorderColor(GameTooltip:GetBackdropBorderColor())
	gldrTooltip:SetBackdropColor(GameTooltip:GetBackdropColor())
	gldrTooltip:SetScale(GameTooltip:GetScale())
	-- assign the header font
	gldrTooltip:SetHeaderFont(headerFont)
	
	self.tooltip = gldrTooltip
	-- populate the tooltip and display it
	LDBObj.ShowTooltip()
end

--- This function hides the tooltip if the mouse cursor leaves the DataBroker object
-- @usage for internal use only
function LDBObj:OnLeave()
	gldrTooltip:Hide()
	QTC:Release(gldrTooltip) -- LibQTipClick
	gldrTooltip = nil
end

--- This function creates and shows the tooltip
-- @usage for internal use only
function LDBObj:ShowTooltip()
	gldrTooltip:AddHeader("Guilder")
	gldrTooltip:AddLine(" ")
	gldrTooltip:AddLine(BLIZZARD..L["Guild Officers"])
	for name in LG:GetIterator("NAME", true) do
		if Guilder:IsOfficer(name) then
			oAlt = Guilder:FindOfficerAlt(name)
			if LG:IsMemberOnline(name) then
				gldrTooltip:AddLine(GREEN..name)
			elseif oAlt ~= nil then
				gldrTooltip:AddLine(YELLOW..oAlt)
			elseif Guilder.db.profile.info.showOffline then
				gldrTooltip:AddLine(RED..name)
			end
		end
	end
	if Guilder.db.profile.info.showModules then
	    gldrTooltip:AddLine(" ")
	    gldrTooltip:AddLine(BLIZZARD..L["Modules"])
	    local modules = {}
		for k, v in Guilder:IterateModules() do
			if v:IsEnabled() then
				table.insert(modules, k.." "..GREEN..L["Enabled"])
			else
				table.insert(modules, k.." "..RED..L["Disabled"])
			end
		end
		table.sort(modules)
		for i, v in ipairs(modules) do
			gldrTooltip:AddLine(v)
		end
	end
	if Guilder.db.profile.info.showHint then
	    gldrTooltip:AddLine(" ")
		gldrTooltip:AddLine(GREEN..L["Hint"]..L["TOOLTIP_HINT"])
	end
	gldrTooltip:Show()
end
